from typing import Dict

from .name import NameFieldGenerator
from .schema import SchemaFieldGenerator
from .base import FieldGenerator
from .number import NumberFieldGenerator
from .date import DateFieldGenerator
from .country import CountryFieldGenerator
from .reference import *

generators: Dict[str, FieldGenerator] = {
    "number": NumberFieldGenerator,
    "date": DateFieldGenerator,
    "country": CountryFieldGenerator,
    "name": NameFieldGenerator
}


def getGeneratorFromFieldType(field_type: str, field_options: Dict[str, str]) -> FieldGenerator:
    Generator = generators.get(field_type, SchemaFieldGenerator)

    return Generator(field_options, field_type)


def get_value_by_attr_name(attr_name: str, schema_name: str, res: Dict = {}, iter_res={}):
    schema = iter_res.get(schema_name, res.get(schema_name))
    if schema is None:
        raise Exception(
            f"Cannot find schema: '{schema_name}' in current result")

    entries = list(filter(
        lambda header: header[1][0] == attr_name, enumerate(schema[0])))

    if len(entries) == 0:
        raise Exception(f"Cannot find attribute: '{attr_name}'")

    entry = entries[0]

    # check if the entry is a schema field generator
    generator = getattr(entry[1][1], "__self__", None)
    if generator is not None and isinstance(generator, SchemaFieldGenerator):
        # find the referenced schema
        return generator.schema

    return schema[1][-1][entry[0]]



from random import random
from typing import Dict, Tuple, List
import os
import csv

from .field_generators.base import *
from .config import loadConfig


def main(file: str, outdir: str):
    """The main entry point for the generate command

        Params:
            file (str): the validated file path
            outdir (str): the path to the 'out'-directory, where data should be stored
    """
    config = loadConfig(file)
    data = generateByConfig(config)

    # write data to disk
    for key in data.keys():
        path = os.path.join(outdir, f"{key}.csv")

        with open(path, "w", newline='', encoding='utf-8') as file:
            writer = csv.writer(file, dialect='excel')
            schema = data.get(key)
            headers = schema[0]

            # write header
            writer.writerow(map(lambda header: header[0], headers))
            # write content
            writer.writerows([[(None if entry is None else headers[index][1](entry)) for (
                index, entry) in enumerate(row)] for row in schema[1]])


def generateByConfig(config: Dict) -> Dict[str, Tuple[Tuple[Tuple], List]]:
    """Generates the data based on the configuration file

        Params:
            config (Dict): the validated configuration file

        Returns:
            data (Dict): the generated data by name and rows
    """

    res: Dict[str, Tuple[Tuple[Tuple], List]] = {}

    generateBaseData(config, res)
    generateFromSteps(config, res)

    return res


def generateBaseData(config: Dict, res: Dict):
    base: Dict[str, int] | None = config.get("base")
    if base is None:
        return

    for (schema_name, amount) in base.items():
        generate_schema(schema_name, config, {}, res, amount=amount)


def generateFromSteps(config: Dict, res: Dict):
    steps: Dict[str, Dict] = config.get("steps")
    if steps is None:
        return

    initial_step = config["initial"]
    amount = config["amount"]

    for _ in range(amount):
        step: Dict = steps.get(initial_step)
        iter_res = {}

        while step is not None:
            generate: str | Dict = step.get("generate")

            # choose what schema to generate
            if type(generate) is str:
                schema_name = generate
            else:
                rand = random()
                total = 0
                schema_name = None

                for (key, value) in generate.items():
                    if (value + total) >= rand:
                        schema_name = key
                        break
                    total += value

            if schema_name is None:
                step = None
                continue

            generate_schema(schema_name, config, res, iter_res)

            step = steps.get(step.get("next"))

        # add generated step data to complete result
        for key, value in iter_res.items():
            if res.get(key) is None:
                res[key] = (value[0], [*value[1]])
            else:
                res[key][1].extend(value[1])


def generate_schema(schema_name: str, config: Dict, res: Dict, iter_res: Dict = {}, amount: int = 1):
    schema: Dict = config["schemas"][schema_name]

    # initialize schema entry if it is not present in result yet
    if iter_res.get(schema_name) is None:
        def map_schema_key(key: str) -> str | Tuple:
            # TODO: return correct val_to_str method
            return (key, schema[key].val_to_str)

        iter_res[schema_name] = (
            ((f"{schema_name.lower()}_id", str), *map(map_schema_key, schema.keys())), [])

    id = 1 if res.get(schema_name) is None else len(
        res[schema_name][1]) + 1

    for _ in range(amount):
        # initialize schema values with id
        schema_entries = iter_res[schema_name][1]
        schema_entries.append([id])

        # generate schema_values
        for field in schema.values():
            if isinstance(field, FieldGenerator):
                value = field.generate(
                    schema_name=schema_name, res=res, iter_res=iter_res)
            else:
                raise Exception(f"Unknown field:\n{field}")

            schema_entries[-1].append(value)

        id += 1



from random import randint

from .base import FieldGenerator

# Pulled from https://github.com/umpirsky/country-list
countries = ("Afghanistan",
             "Åland Islands",
             "Albania",
             "Algeria",
             "American Samoa",
             "Andorra",
             "Angola",
             "Anguilla",
             "Antarctica",
             "Antigua & Barbuda",
             "Argentina",
             "Armenia",
             "Aruba",
             "Australia",
             "Austria",
             "Azerbaijan",
             "Bahamas",
             "Bahrain",
             "Bangladesh",
             "Barbados",
             "Belarus",
             "Belgium",
             "Belize",
             "Benin",
             "Bermuda",
             "Bhutan",
             "Bolivia",
             "Bosnia & Herzegovina",
             "Botswana",
             "Bouvet Island",
             "Brazil",
             "British Indian Ocean Territory",
             "British Virgin Islands",
             "Brunei",
             "Bulgaria",
             "Burkina Faso",
             "Burundi",
             "Cambodia",
             "Cameroon",
             "Canada",
             "Cape Verde",
             "Caribbean Netherlands",
             "Cayman Islands",
             "Central African Republic",
             "Chad",
             "Chile",
             "China",
             "Christmas Island",
             "Cocos (Keeling) Islands",
             "Colombia",
             "Comoros",
             "Congo - Brazzaville",
             "Congo - Kinshasa",
             "Cook Islands",
             "Costa Rica",
             "Côte d’Ivoire",
             "Croatia",
             "Cuba",
             "Curaçao",
             "Cyprus",
             "Czechia",
             "Denmark",
             "Djibouti",
             "Dominica",
             "Dominican Republic",
             "Ecuador",
             "Egypt",
             "El Salvador",
             "Equatorial Guinea",
             "Eritrea",
             "Estonia",
             "Eswatini",
             "Ethiopia",
             "Falkland Islands",
             "Faroe Islands",
             "Fiji",
             "Finland",
             "France",
             "French Guiana",
             "French Polynesia",
             "French Southern Territories",
             "Gabon",
             "Gambia",
             "Georgia",
             "Germany",
             "Ghana",
             "Gibraltar",
             "Greece",
             "Greenland",
             "Grenada",
             "Guadeloupe",
             "Guam",
             "Guatemala",
             "Guernsey",
             "Guinea",
             "Guinea-Bissau",
             "Guyana",
             "Haiti",
             "Heard & McDonald Islands",
             "Honduras",
             "Hong Kong SAR China",
             "Hungary",
             "Iceland",
             "India",
             "Indonesia",
             "Iran",
             "Iraq",
             "Ireland",
             "Isle of Man",
             "Israel",
             "Italy",
             "Jamaica",
             "Japan",
             "Jersey",
             "Jordan",
             "Kazakhstan",
             "Kenya",
             "Kiribati",
             "Kuwait",
             "Kyrgyzstan",
             "Laos",
             "Latvia",
             "Lebanon",
             "Lesotho",
             "Liberia",
             "Libya",
             "Liechtenstein",
             "Lithuania",
             "Luxembourg",
             "Macao SAR China",
             "Madagascar",
             "Malawi",
             "Malaysia",
             "Maldives",
             "Mali",
             "Malta",
             "Marshall Islands",
             "Martinique",
             "Mauritania",
             "Mauritius",
             "Mayotte",
             "Mexico",
             "Micronesia",
             "Moldova",
             "Monaco",
             "Mongolia",
             "Montenegro",
             "Montserrat",
             "Morocco",
             "Mozambique",
             "Myanmar (Burma)",
             "Namibia",
             "Nauru",
             "Nepal",
             "Netherlands",
             "New Caledonia",
             "New Zealand",
             "Nicaragua",
             "Niger",
             "Nigeria",
             "Niue",
             "Norfolk Island",
             "North Korea",
             "North Macedonia",
             "Northern Mariana Islands",
             "Norway",
             "Oman",
             "Pakistan",
             "Palau",
             "Palestinian Territories",
             "Panama",
             "Papua New Guinea",
             "Paraguay",
             "Peru",
             "Philippines",
             "Pitcairn Islands",
             "Poland",
             "Portugal",
             "Puerto Rico",
             "Qatar",
             "Réunion",
             "Romania",
             "Russia",
             "Rwanda",
             "Samoa",
             "San Marino",
             "São Tomé & Príncipe",
             "Saudi Arabia",
             "Senegal",
             "Serbia",
             "Seychelles",
             "Sierra Leone",
             "Singapore",
             "Sint Maarten",
             "Slovakia",
             "Slovenia",
             "Solomon Islands",
             "Somalia",
             "South Africa",
             "South Georgia & South Sandwich Islands",
             "South Korea",
             "South Sudan",
             "Spain",
             "Sri Lanka",
             "St. Barthélemy",
             "St. Helena",
             "St. Kitts & Nevis",
             "St. Lucia",
             "St. Martin",
             "St. Pierre & Miquelon",
             "St. Vincent & Grenadines",
             "Sudan",
             "Suriname",
             "Svalbard & Jan Mayen",
             "Sweden",
             "Switzerland",
             "Syria",
             "Taiwan",
             "Tajikistan",
             "Tanzania",
             "Thailand",
             "Timor-Leste",
             "Togo",
             "Tokelau",
             "Tonga",
             "Trinidad & Tobago",
             "Tunisia",
             "Turkey",
             "Turkmenistan",
             "Turks & Caicos Islands",
             "Tuvalu",
             "U.S. Outlying Islands",
             "U.S. Virgin Islands",
             "Uganda",
             "Ukraine",
             "United Arab Emirates",
             "United Kingdom",
             "United States",
             "Uruguay",
             "Uzbekistan",
             "Vanuatu",
             "Vatican City",
             "Venezuela",
             "Vietnam",
             "Wallis & Futuna",
             "Western Sahara",
             "Yemen",
             "Zambia",
             "Zimbabwe")

count = len(countries) - 1


class CountryFieldGenerator(FieldGenerator):
    def generate(self, **_):
        return countries[randint(0, count)]



from random import randint

from .base import FieldGenerator

# Pulled from https://www.usna.edu/Users/cs/roche/courses/s15si335/proj1/files.php%3Ff=names.txt.html
names = ("Michael",
         "Christopher",
         "Jessica",
         "Matthew",
         "Ashley",
         "Jennifer",
         "Joshua",
         "Amanda",
         "Daniel",
         "David",
         "James",
         "Robert",
         "John",
         "Joseph",
         "Andrew",
         "Ryan",
         "Brandon",
         "Jason",
         "Justin",
         "Sarah",
         "William",
         "Jonathan",
         "Stephanie",
         "Brian",
         "Nicole",
         "Nicholas",
         "Anthony",
         "Heather",
         "Eric",
         "Elizabeth",
         "Adam",
         "Megan",
         "Melissa",
         "Kevin",
         "Steven",
         "Thomas",
         "Timothy",
         "Christina",
         "Kyle",
         "Rachel",
         "Laura",
         "Lauren",
         "Amber",
         "Brittany",
         "Danielle",
         "Richard",
         "Kimberly",
         "Jeffrey",
         "Amy",
         "Crystal",
         "Michelle",
         "Tiffany",
         "Jeremy",
         "Benjamin",
         "Mark",
         "Emily",
         "Aaron",
         "Charles",
         "Rebecca",
         "Jacob",
         "Stephen",
         "Patrick",
         "Sean",
         "Erin",
         "Zachary",
         "Jamie",
         "Kelly",
         "Samantha",
         "Nathan",
         "Sara",
         "Dustin",
         "Paul",
         "Angela",
         "Tyler",
         "Scott",
         "Katherine",
         "Andrea",
         "Gregory",
         "Erica",
         "Mary",
         "Travis",
         "Lisa",
         "Kenneth",
         "Bryan",
         "Lindsey",
         "Kristen",
         "Jose",
         "Alexander",
         "Jesse",
         "Katie",
         "Lindsay",
         "Shannon",
         "Vanessa",
         "Courtney",
         "Christine",
         "Alicia",
         "Cody",
         "Allison",
         "Bradley",
         "Samuel")

count = len(names) - 1


class NameFieldGenerator(FieldGenerator):
    def generate(self, **_):
        return names[randint(0, count)]



from datetime import date, timedelta
from schema import Or, Schema, Use, Optional
from random import randint

from .reference import to_ref
from .base import FieldGenerator


class DateFieldGenerator(FieldGenerator):
    option_schema = Schema({
        "min": Or(Use(to_ref), Use(date.fromisoformat)),
        "max": Or(Use(to_ref), Use(date.fromisoformat)),
        Optional("delta_min", default=0): Use(int),
        Optional("delta_max"): Use(int),
        Optional("delta"): Use(int)
    })

    def validate(self):
        self.options = self.option_schema.validate(self.options)

        delta = self.options.get("delta") is not None
        delta_min = self.options.get("delta_min") is not None
        delta_max = self.options.get("delta_max") is not None

        if delta and (delta_min or delta_max):
            raise Exception(
                f"Provide either the 'delta' option or the options 'delta_min' and 'delta_max'")

    def generate(self, **kwargs):
        options = self.prepare_options(**kwargs)
        min = options["min"]
        max = options["max"]

        if max < min:
            raise Exception(
                f"Option 'max' ({max}) must be equal or larger than 'min' ({min})")

        delta = options.get("delta")

        if delta is not None:
            add_days = delta
        else:
            delta_min = options["delta_min"]
            delta_max = options.get("delta_max", delta_min)

            if delta_max < delta_min:
                raise Exception(
                    f"'delta_max' ({delta_max}) must be equal or larger than 'delta_min' ({delta_min})")

            add_days = randint(delta_min, delta_max)

        day_difference = (max - min).days
        value = min + timedelta(days=randint(0, day_difference) + add_days)

        return value



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Tuple

from .reference import ReferenceFieldGenerator


class FieldGenerator(object, metaclass=ABCMeta):
    def __init__(self, options: Dict[str, str] = {}, type: str = None):
        super(FieldGenerator, self).__init__()
        self.options = options
        self.type = type
        self.validate()

    @abstractmethod
    def generate(self):
        raise NotImplementedError()

    def validate(self):
        pass

    def val_to_str(self, value: Any) -> str:
        return str(value)

    def prepare_options(self, schema_name: str = None, res: Dict[str, Tuple[Tuple, List[List]]] = {}, iter_res={}):
        options = self.options.copy()
        for (key, value) in options.items():
            if not isinstance(value, ReferenceFieldGenerator):
                continue

            schema = iter_res.get(value.type, res.get(value.type))
            if schema is None:
                raise Exception(
                    f"Cannot find schema: '{value.type}' in current result")

            entry = list(filter(
                lambda header: header[1][0] == value.options["field_attr"], enumerate(schema[0])))[0]

            options[key] = schema[1][-1][entry[0]]

        return options



"""
This is the "load_config" module.

The load config helps to load and validate the sledo configuration files.
"""

import importlib
from io import TextIOWrapper
from os import path
import pathlib
import sys
from schema import And, Or, Schema, Optional, SchemaError, Use
from typing import Dict
import yaml

from .field_generators.base import FieldGenerator
from .field_generators.reference import ReferenceFieldGenerator
from .field_generators.schema import SchemaFieldGenerator
from .field_generators import getGeneratorFromFieldType
from sledo.generate import field_generators


class ConfigurationSchema(Schema):
    def validate(self, data, _is_config_schema=True):
        data = super(ConfigurationSchema, self).validate(
            data, _is_config_schema=False)

        if not _is_config_schema:
            return data

        steps: Dict[str, Dict[str, str]] = data.get("steps")
        initial: str = data.get("initial")
        base: Dict[str, str] = data.get("base")
        schemas: Dict[str, Dict[str, FieldGenerator]] = data["schemas"]

        # validate base data
        if base is not None:
            for schema_name in base.keys():
                if schemas.get(schema_name) is None:
                    raise SchemaError(f"Missing schema: '{schema_name}'")

        if initial is not None and steps is None:
            raise SchemaError(
                f"Missing required field 'steps', when using 'initial'")

        if steps is not None:
            # validate initial step
            if steps.get(data["initial"]) is None:
                raise SchemaError(f"Missing step: '{data['initial']}'")

            # validate next steps
            for (key, step) in steps.items():
                next = step.get("next")
                if next is not None and steps.get(next) is None:
                    raise SchemaError(f"Missing step: '{next}'")

                generate: str | Dict = step.get("generate")

                if type(generate) is str:
                    if schemas.get(generate) is None:
                        raise SchemaError(f"Missing schema: '{generate}'")
                else:
                    total_prob = 0
                    for (schema, prob) in generate.items():
                        if schemas.get(schema) is None:
                            raise SchemaError(f"Missing schema: '{schema}'")

                        total_prob += prob

                    if total_prob > 1:
                        raise SchemaError(
                            f"The total probability must not be more than 1 at step: '{key}'")

        # validate properties
        for schema in schemas.values():
            for value in schema.values():
                if isinstance(value, SchemaFieldGenerator):
                    # check if Schema exists
                    ref = schemas.get(value.type)
                    if ref is None:
                        raise SchemaError(f"Missing schema: '{value.type}'")
                elif isinstance(value, FieldGenerator):
                    for value in value.options.values():
                        if isinstance(value, ReferenceFieldGenerator):
                            ref_field = schema.get(value.options["field"])
                            if ref_field is None:
                                raise SchemaError(
                                    f"Missing attribute: '{value.options['field']}'")

                            if not isinstance(ref_field, SchemaFieldGenerator):
                                raise SchemaError(
                                    f"Attribute '{value.options['field']}' must be a schema reference!")

                            value.type = ref_field.type
                            ref_schema = schemas.get(ref_field.type)
                            ref_schema_field = ref_schema.get(
                                value.options["field_attr"])

                            if ref_schema_field is None:
                                raise SchemaError(
                                    f"Attribute '{value.options['field_attr']}' does not exist in schema {ref_field.type}")

        return data


def to_generator(raw: str | dict):
    is_str = type(raw) is str
    field_type = raw if is_str else raw["type"]
    field_options = {} if is_str else {
        k: raw[k] for k in raw if k != "type"
    }

    generator = getGeneratorFromFieldType(field_type, field_options)

    return generator


# The schema of a valid configuration file
schema = ConfigurationSchema({
    Optional('base'): {
        str: And(Use(int), lambda x: x >= 0)
    },
    Optional('amount'): Use(int),
    Optional('generators'): {
        str: str
    },
    Optional('initial'): str,
    Optional('steps'): {
        str: {
            "generate": Or(str, {
                str: And(Use(float), lambda x: 0 <= x <= 1)
            }),
            Optional("next"): str
        }
    },
    'schemas': {
        str: {
            str: And(Or(str, {
                "type": str,
                Optional(str): str
            }), Use(to_generator))
        }
    }
})


def validateConfig(config: Dict, MODULE_DIR=None):
    # load generators before everything else is validated
    generators: Dict[str, str] = config.get("generators", {})

    if len(generators) > 0:
        if MODULE_DIR is None:
            raise Exception(
                "Cannot import generators without directory path!")
        sys.path.append(MODULE_DIR)

    # load generators
    for (name, rel_path) in generators.items():
        module_name = pathlib.Path(path.join(MODULE_DIR, rel_path)).stem
        module = importlib.import_module(module_name)

        field_generators.generators[name] = getattr(module, name)

    return schema.validate(config)


def loadConfig(file: TextIOWrapper) -> Dict:
    """
    Loads the configuration file and validates its contents

        Params:
            file (str): the file path of the configuration file

        Returns:
            config (Dict): the loaded configuration file
    """

    # load and parse the configuration file
    config: Dict = yaml.load(file, Loader=yaml.BaseLoader)

    dir_path = path.dirname(path.realpath(file.name))

    # validate and transform the loaded configuration file
    return validateConfig(config, dir_path)



import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="sledo",
    version="0.2.1",
    description="A simple demodata generator for the command-line.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/luxbe/sledo",
    author="luxbe, Jakob, Ammar, Melli, Deniz, Larissa",
    author_email="luxbe@tutanota.com, jakobjonathan@aol.com, ammar@begovic.eu, melina.petersen99@web.de, denizcan.1999er@gmail.com, larissac0504@gmail.com",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=[
        'click',
        'pyyaml',
        'schema'
    ],
    entry_points={
        "console_scripts": [
            "sledo=sledo:cli"
        ]
    },
)
